#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#ifndef USE_NETLIB
#include <net/ethernet.h>
#include <netinet/in.h>
#ifdef __linux__
#include <netinet/ether.h>
#endif
#include <arpa/inet.h>
#else
#include <netlib.h>
#endif

#include "pktlib.h"
#include "pktbuf.h"

struct macaddr_entry {
  struct ether_addr macaddr;
  pktif_t pktif;
};

#define MACADDR_TABLE_ALLNUM  (1024*4)
#define MACADDR_TABLE_LINENUM (16*4)
#define MACADDR_TABLE_COLNUM  (MACADDR_TABLE_ALLNUM / MACADDR_TABLE_LINENUM)

#define MACADDR_TABLE_HASH(macaddr) \
	( \
		((unsigned int)((unsigned char *)(macaddr))[2] << 24) | \
		((unsigned int)((unsigned char *)(macaddr))[3] << 16) | \
		((unsigned int)((unsigned char *)(macaddr))[4] <<  8) | \
		((unsigned int)((unsigned char *)(macaddr))[5] <<  0) \
	)
#define MACADDR_TABLE_INDEX(macaddr) \
	(MACADDR_TABLE_HASH(macaddr) % MACADDR_TABLE_LINENUM)

static struct macaddr_entry macaddr_table[MACADDR_TABLE_LINENUM][MACADDR_TABLE_COLNUM];
static unsigned int macaddr_hash[MACADDR_TABLE_LINENUM][MACADDR_TABLE_COLNUM];

int macaddr_table_init()
{
  memset(macaddr_table, 0, sizeof(macaddr_table));
  memset(macaddr_hash, 0, sizeof(macaddr_hash));
  return 0;
}

int macaddr_entry_dump()
{
  struct macaddr_entry *entry;
  int index, i;

  printf("---- MAC Address Table ----\n");

  for (index = 0; index < MACADDR_TABLE_LINENUM; index++) {
    for (i = 0; i < MACADDR_TABLE_COLNUM; i++) {
      entry = &macaddr_table[index][i];
      if (entry->pktif) {
	printf("LINE[%04d]:\tMAC Address: %s (Interface: %s)\n", index,
	       ether_ntoa(&entry->macaddr), pktif_get_name(entry->pktif));
      }
    }
  }
  fflush(stdout);

  return 0;
}

struct macaddr_entry *macaddr_entry_search(char macaddr[])
{
  struct macaddr_entry *entry;
  unsigned int hash;
  int index, i;

  hash = MACADDR_TABLE_HASH(macaddr);
  index = MACADDR_TABLE_INDEX(macaddr);

  for (i = 0; i < MACADDR_TABLE_COLNUM; i++) {
    if (macaddr_hash[index][i] == hash) {
      entry = &macaddr_table[index][i];
      if (!memcmp(macaddr, &entry->macaddr, ETHER_ADDR_LEN))
	return entry;
    }
  }

  return NULL;
}

struct macaddr_entry *macaddr_entry_create(char macaddr[])
{
  struct macaddr_entry *entry;
  int index, i;

  index = MACADDR_TABLE_INDEX(macaddr);

  for (i = 0; i < MACADDR_TABLE_COLNUM; i++) {
    entry = &macaddr_table[index][i];
    if (entry->pktif == NULL) {
      memcpy(&entry->macaddr, macaddr, ETHER_ADDR_LEN);
      macaddr_hash[index][i] = MACADDR_TABLE_HASH(macaddr);
      return entry;
    }
  }

  return NULL;
}

struct macaddr_entry *macaddr_entry_add(char macaddr[], pktif_t pktif)
{
  struct macaddr_entry *entry;

  entry = macaddr_entry_search(macaddr);
  if (entry == NULL)
    entry = macaddr_entry_create(macaddr);
  if (entry && (entry->pktif != pktif)) {
    entry->pktif = pktif;
    macaddr_entry_dump();
  }

  return entry;
}

void usage()
{
  printf("[-l <lost_rate>] [-n <noise_rate>] [-d <delay_rate>] \\\n");
  printf("[-i <interface>]...\n");
  exit(0);
}

struct pktbuf_option {
  pktif_t rcvif;
};

int main(int argc, char *argv[])
{
  pktif_t pktif, p;
  int r, ch, t, lost_rate = 0, noise_rate = 0, delay_rate = 0;
  pktbuf_t pktbuf;
  int type;
  struct header_param param;
  struct macaddr_entry *entry;
  pktbuf_t packets = NULL, waiting = NULL;

  while ((ch = getopt(argc, argv, "l:n:d:i:")) != -1) {
    switch (ch) {
    case 'l':
      lost_rate = atoi(optarg);
      break;
    case 'n':
      noise_rate = atoi(optarg);
      break;
    case 'd':
      delay_rate = atoi(optarg);
      break;
    case 'i':
      /* 󥿡ե򥪡ץ󤹤(ȯѥåȤϼʤ) */
      pktif_open(optarg, PKTIF_OPEN_FLAG_RECV_NOTSENT, 0);
      break;
    case '?':
    default:
      usage();
    }
  }

  srand(time(NULL));

  pktbuf_init(sizeof(struct pktbuf_option));

  macaddr_table_init();

  while (1) {
    if (packets == NULL) {
      t = waiting ? 3000000: -1;
      pktif = pktlib_iflist_select(t); /* ʣ󥿡եԤ */
      if (pktif) {
	pktbuf = pktbuf_recv(pktif, 0); /* ѥåȤ */
	if (pktbuf)
	  ((struct pktbuf_option *)pktbuf_get_option(pktbuf))->rcvif = pktif;
	pktbuf_enqueue(&packets, pktbuf);
      }
      pktbuf_enqueue(&packets, waiting);
      waiting = NULL;
      continue;
    }

    pktbuf = pktbuf_dequeue(&packets);
    pktif = ((struct pktbuf_option *)pktbuf_get_option(pktbuf))->rcvif;

    if ((delay_rate > 0) && ((rand() % 100) < delay_rate)) {
      /* ꤵ줿Ψٱ䤵 */
      pktbuf_enqueue(&waiting, pktbuf);
      continue;
    }

    if ((lost_rate > 0) && ((rand() % 100) < lost_rate))
      goto discard; /* ꤵ줿ΨǥѥåȤѴ */

    if ((noise_rate > 0) && ((rand() % 100) < noise_rate)) {
      /* ꤵ줿ΨǥΥ */
      pktbuf_get_header(pktbuf)[rand() % pktbuf_get_size(pktbuf)] ^=
	(1 << (rand() % 8));
    }

    type = pktif_get_linktype(pktif);
    if (type != DLT_EN10MB)
      goto discard;

    r = pktbuf_analyze_ethernet(pktbuf, &param);
    if (r < 1)
      goto discard;

    /* MACɥ쥹ؽ */
    macaddr_entry_add(param.ethernet.src_macaddr, pktif);

    /* ѥåȤ */
    entry = macaddr_entry_search(param.ethernet.dst_macaddr);
    if ((param.ethernet.dst_macaddr[0] & 1) || (entry == NULL)) {
      /* ֥ɥ㥹ȤݡȤ */
      for (p = pktlib_iflist_get_list(); p; p = pktif_get_next(p)) {
	if (p != pktif)
	  pktbuf_send(p, pktbuf_clone(pktbuf));
      }
    } else {
      pktbuf = pktbuf_send(entry->pktif, pktbuf);
    }

discard:
    pktbuf_destroy(pktbuf);
  }

  return 0;
}
